import os,sys,getopt

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        pass
    try:
        import unicodedata
        unicodedata.numeric(s)
        return True
    except (TypeError, ValueError):
        pass
    return False

def all_numbers(numlist):
    for s in numlist:
        if is_number(s)==False:
            return False
    return True

def parse_id(inpid):
    l,ids=[],[]
    tmp=inpid.strip('"').split()
    for i in tmp:
        l.extend(i.split(','))
    for i in l:
        if '-' in i:
            a,b=i.split('-')
            for j in range(int(a),int(b)+1):
                ids.append(j)
        else:
            ids.append(int(i))
    return ids

def int_occ_num(occ):
    if 0<=occ<=0.5:
        return 0
    elif 0.5<occ<1.5:
        return 1
    elif 1.5<=occ:
        return 2
    else:
        sys.exit(f'Error occupation number: {occ}')

def find(mylist,a,b):
    ia,ib=None,None
    for i in range(len(mylist)):
        if a==mylist[i]:
            ia=i
        elif b==mylist[i]:
            ib=i
    return ia,ib

def readfch_r(fch):
    na,nb,norb,occ=None,None,None,[]
    with open(fch,'r') as read:
        flag=False
        for line in read.readlines():
            if 'Number of alpha electrons' in line:
                na=int(line.split()[-1])
            elif 'Number of beta electrons' in line:
                nb=int(line.split()[-1])
            elif 'Number of basis functions' in line:
                norb=int(line.split()[-1])
            elif 'Alpha Orbital Energies' in line:
                flag=True
                continue
            elif 'Alpha MO coefficients' in line:
                break
            if flag:
                for i in line.split():
                    occ.append(int_occ_num(float(i)))
    return na,nb,norb,occ

def sort_r(fch,inpid):
    na,nb,norb,occ=readfch_r(fch); tosort=[i for i in range(1,1+norb)]

    active=parse_id(inpid); no=len(active); ne=0
    for i in active:
        ne+=occ[i-1]
    print(f'CASSCF({no}o,{ne}e)')
    nocc=min(na,nb); nopen=abs(na-nb); nvir=norb-nocc-nopen

    active_occ,active_open,active_vir=[],[],[]
    for i in active:
        if occ[i-1]>=1.5:
            active_occ.append(i)
        elif 0.5<=occ[i-1]<1.5:
            active_open.append(i)
        else:
            active_vir.append(i)
    print(active_occ,active_open,active_vir)
    sorting=[]
#   for i in range(1,1+nocc):
#       if i not in active_occ:
#           sorting.append(i)
#   for i in range(1+nocc,1+nocc+nopen):
#       if i not in active_open:
#           sorting.append(i)
#   for i in range(1,1+nocc):
#       if i in active_occ:
#           sorting.append(i)
#   for i in range(1+nocc,1+nocc+nopen):
#       if i in active_open:
#           sorting.append(i)
#   for i in range(1+nocc+nopen,1+norb):
#       if i in active_vir:
#           sorting.append(i)
#   for i in range(1+nocc+nopen,1+norb):
#       if i not in active_vir:
#           sorting.append(i)
    for i in range(norb):
        if 1+i not in active:
            if occ[i]>=1.5:
                sorting.append(1+i)
    sorting.extend(active_occ)
    sorting.extend(active_open)
    sorting.extend(active_vir)
    for i in range(norb):
        if 1+i not in sorting:
            sorting.append(1+i)

    ex=[]
    for i in range(norb):
        if tosort[i]==sorting[i]:
            pass
        else:
            ia,ib=find(tosort,tosort[i],sorting[i])
            tmp=tosort[ia]; tosort[ia]=tosort[ib]; tosort[ib]=tmp
            ex.append([1+ia,1+ib])
    # print(ex)
    return ex

def readfch_u(fch):
    na,nao,aocc=None,None,[]
    nb,nbo,bocc=None,None,[]
    with open(fch,'r') as read:
        flag=False
        for line in read.readlines():
            if 'Number of alpha electrons' in line:
                na=int(line.split()[-1])
            elif 'Number of beta electrons' in line:
                nb=int(line.split()[-1])
            elif 'Number of basis functions' in line:
                nao=int(line.split()[-1])
                nbo=int(line.split()[-1])
    aocc=[1 for i in range(na)]
    aocc.extend([0 for i in range(nao-na)])
    bocc=[1 for i in range(nb)]
    bocc.extend([0 for i in range(nbo-nb)])
    return na,nb,nao,nbo,aocc,bocc

def sort_u(fch,inpid):
    na,nb,nao,nbo,aocc,bocc=readfch_u(fch)

    active=parse_id(inpid); tosort=[i for i in range(1,1+nao+nbo)]

    active=parse_id(inpid)
    active_aocc,active_avir=[],[]
    active_bocc,active_bvir=[],[]
    for i in active:
        if 0<i<=na:
            active_aocc.append(i)
        elif na<i<=nao:
            active_avir.append(i)
        elif nao<i<=nao+nb:
            active_bocc.append(i)
        elif nao+nb<i<=nao+nbo:
            active_bvir.append(i)
    active_nao=len(active_aocc)+len(active_avir)
    active_nbo=len(active_bocc)+len(active_bvir)
    print(f'CASSCF(({len(active_aocc)+len(active_avir)},{len(active_bocc)+len(active_bvir)}),({len(active_aocc)},{len(active_bocc)}))')

    sorting=[]
    # sort alpha occupied orbitals
    for i in range(1,1+na):
        if i not in active_aocc:
            sorting.append(i)
    for i in range(1,1+na):
        if i in active_aocc:
            sorting.append(i)
    # sort alpha virtual orbitals
    for i in range(1+na,1+nao):
        if i in active_avir:
            sorting.append(i)
    for i in range(1+na,1+nao):
        if i not in active_avir:
            sorting.append(i)
    # sort beta occupied orbitals
    for i in range(1+nao,1+nao+nb):
        if i not in active_bocc:
            sorting.append(i)
    for i in range(1+nao,1+nao+nb):
        if i in active_bocc:
            sorting.append(i)
    # sort beta virtual orbitals
    for i in range(1+nao+nb,1+nao+nbo):
        if i in active_bvir:
            sorting.append(i)
    for i in range(1+nao+nb,1+nao+nbo):
        if i not in active_bvir:
            sorting.append(i)
    
    ex=[]
    for i in range(nao+nbo):
        if tosort[i]==sorting[i]:
            pass
        else:
            ia,ib=find(tosort,tosort[i],sorting[i])
            tmp=tosort[ia]; tosort[ia]=tosort[ib]; tosort[ib]=tmp
            ex.append([1+ia,1+ib])
    return ex

def read_fragment_setting(inp):
    system_charge,system_mult=0,1
    nfrag=0; frag_charges,frag_mults=[],[]
    atms=[]
    with open(inp,'r') as read:
        flag=False
        for line in read.readlines():
            l=line.split()
            for i in l:
                if 'guess' in i.lower() and 'fragment' in i.lower():
                    nfrag=int(i.split('=')[-1].rstrip(')'))
            if nfrag!=0:
                if len(l)==2*nfrag+2 and all_numbers(l):
                    system_charge=int(l[0]); system_mult=int(l[1])
                    for i in range(nfrag):
                        frag_charges.append(int(l[2+2*i]))
                        frag_mults.append(int(l[3+2*i]))
                    flag=True
            if flag and len(l)==4:
                atms.append([l[0].split('(')[0],
                             int(l[0].split('=')[-1].strip(')')),
                             float(l[1]),float(l[2]),float(l[3])])
    return nfrag,frag_charges,frag_mults,atms

def generate_fragment_input(inp):
    nfrag,frag_charges,frag_mults,atms=read_fragment_setting(inp)
    global method
    for i in range(nfrag):
        with open(f'frag{1+i}.gjf','w') as gen:
            gen.write(f'''\
%nproc=16
%mem=48GB
%chk=frag{1+i}.chk
#p {method} scf(maxcyc=512,xqc) stable=opt
   nosymm int=nobasistransform 5d 7f

generated by exmo.py

{frag_charges[i]} {frag_mults[i]}
''')
            for atm in atms:
                if atm[1]==1+i:
                    gen.write(f'{atm[0]:>4}')
                    for j in range(3):
                        gen.write(f'  {atm[2+j]:>16.8f}')
                    gen.write('\n')
            gen.write('\n\n')
    return

def generate_fragment_guess(inp,output='tosort.fchk'):
    nfrag,frag_charges,frag_mults,atms=read_fragment_setting(inp)
    with open('puzzle.sh','w') as gen:
        gen.write(f'''\
#!/bin/bash

python kttk_mixer.py >> tmp << EOF
''')
        for i in range(nfrag):
            gen.write(f'frag{1+i}.fchk\n')
        gen.write('\n')
        for i in range(nfrag):
            if frag_mults[i]<0:
                gen.write('-1 ')
            else:
                gen.write(' 1 ')
        gen.write(f'''
{output}
EOF
    rm tmp
''')
    os.system('chmod +x puzzle.sh; ./puzzle.sh; rm puzzle.sh')
    read_all_fragment_fch(nfrag)
    return

def read_all_fragment_fch(nfrag):
    nas,nbs,naos,nbos,aoccs,boccs=[],[],[],[],[],[]
    for i in range(nfrag):
        na,nb,nao,nbo,aocc,bocc=readfch_u(f'frag{1+i}.fchk')
        nas.append(na); naos.append(nao); aoccs.append(aocc)
        nbs.append(nb); nbos.append(nbo); boccs.append(bocc)

    nbasis,na,nb=0,0,0
    for i in range(nfrag):
        nbasis+=naos[i]
        na+=nas[i]
        nb+=nbs[i]

    aocc_count=0
    avir_count=na
    bocc_count=nbasis
    bvir_count=nbasis+nb
    print('''
    Fragment wave function information
  -----------------------------------------------------------------------------------
    Fragment  n_basis  n_a_ele  a_occ_orb  a_vir_orb  n_b_ele  b_occ_orb  b_vir_orb
  -----------------------------------------------------------------------------------''')
    for i in range(nfrag):
        print(f'  \
  {1+i:>8}\
  {naos[i]:>7}\
  {nas[i]:>7}\
  {1+aocc_count:>4}-{aocc_count+nas[i]:>4}\
  {1+avir_count:>4}-{avir_count+naos[i]-nas[i]:>4}\
  {nbs[i]:>7}\
  {1+bocc_count:>4}-{bocc_count+nbs[i]:>4}\
  {1+bvir_count:>4}-{bvir_count+nbos[i]-nbs[i]:>4}')
        aocc_count+=nas[i]
        avir_count+=naos[i]-nas[i]
        bocc_count+=nbs[i]
        bvir_count+=nbos[i]-nbs[i]
    print(f'''\
  -----------------------------------------------------------------------------------
      System  {nbasis:>7}  {na:>7}                        {nb:>7}
''')
    return

def exchange_MO(fch,inpid,wave_function_type='r'):
    ex=None
    if wave_function_type=='r':
        ex=sort_r(fch,inpid)
    elif wave_function_type=='u':
        ex=sort_u(fch,inpid)

    global output,method
    mokit=True; mwfn=False
    if mokit:
        with os.popen(f'cp {fch} {output}') as cp:
            tmp=cp.read()
        from mokit.lib.gaussian import permute_orb
        for i in ex:
            permute_orb(output,i[0],i[1])
    elif mwfn:
        with open('exchange_MO.sh','w') as gen:
            gen.write(f'''\
#!/bin/bash

Multiwfn {fch} >> tmp << EOF
6
''')
        for i in ex:
            gen.write(f'''\
29
{i[0]},{i[1]}
''')
        gen.write(f'''\
-1
100
2
7
tmp.fchk
0
q
EOF

rm tmp

    cat >> formatted.gjf << EOF
%chk=sorted.chk
%oldchk=tmp.chk
#p {method}
   nosymm int=nobasistransform 5d 7f
   guess=(read,only) geom=allchk
EOF

    unfchk tmp.fchk
    g16 formatted.gjf

    rm tmp.fchk tmp.chk formatted.gjf formatted.log
    formchk sorted.chk
    mv sorted.fchk {output}
    rm sorted.chk
''')
        with os.popen('chmod +x exchange_MO.sh; ./exchange_MO.sh') as run:
            null=run.read()

def printUsage():
    print('''
    Usage:
      (1) Based on fragments guess.
        python exmo.py -s 1 -i system.gjf -m UHF/def2TZVP
        python exmo.py -s 2 -i system.gjf -p tosort.fch
        python exmo.py -s 3 -i tosort.fch -n "30-40,42,78,80-90" -t UHF -p sorted.fch -m UHF/def2TZVP
      (2) Based on NO wavefunction.
        python exmo.py -s 3 -i tosort.fch -n "30-40,42,78,80-90" -t NO -p sorted.fch -m UHF/def2TZVP

    Options:
      -h: Print this information and exit.
      -i: Input fragment setting or wave function file.
      -n: Number of orbitals to be included in the active space.
      -t: Tpye of wave function.
             NO: (R/RO)NO with orbital energy in .fch replaced by occupation number.
            UHF: UHF wave function.
      -p: Print sorted wave function to file. Default: sorted.fch
      -m: Method/basis settings. Default: UHF/def2TZVP
      -s: Step of calculation.

    Dependencies:
      Multiwfn (run by "Multiwfn")
      Gaussian 16 (run by "g16")
''')
    exit()
    return

def main():
    filename=None; inpid=None; wave_function_type='r'
    global output,method
    output='sorted.fch'; method='UHF/def2TZVP'
    try:
        opts,args=getopt.getopt(sys.argv[1:],'hi:n:t:p:m:s:')
    except getopt.GetoptError:
        printUsage()
        sys.exit(-1)
    for opt,arg in opts:
        if opt=='-h':
            printUsage()
        elif opt=='-i':
            filename=arg
        elif opt=='-n':
            inpid=arg
        elif opt=='-t':
            if arg.lower() in ['no']:
                wave_function_type='r'
            elif arg.lower() in ['uhf']:
                wave_function_type='u'
        elif opt=='-p':
            output=arg
        elif opt=='-m':
            method=arg
        elif opt=='-s':
            step=int(arg)

    # main
    if step==1:
        generate_fragment_input(filename)

    elif step==2:
        generate_fragment_guess(filename,output)

    elif step==3:
        if os.path.exists('exmo_tmp.py'):
            import exmo_tmp
            method=exmo_tmp.method_save
        exchange_MO(filename,inpid,wave_function_type)

if __name__=='__main__':
    main()

